const v4 = require('uuid').v4;
const {
    Before,
    BeforeAll,
    After,
    Given,
    Then,
    When,
} = require('@cucumber/cucumber');
const { expect, parseRelativeDate, oid } = require('./util');
const {
    drgWithAttach,
    env,
    env_nas,
    patient_cpr,
    number_of_relatives,
} = require('dataaccess/util');

const services = new Map(require('../config/services.json'));
const roles = new Map(require('../config/roles.json'));

const newBody = (id, role, endpoint) => ({
    id,
    role,
    endpoint: endpoint || env,
    fields: {},
    'auth-attributes': {},
});

Given('datasæt {string} med felterne', function (name, dataTable) {
    const data = this.parseDataTable(dataTable);
    // If statement tilføjet for at fjerne z i data, for Fibs Newborn registration ikke kan håndtere at der er z i slutningen.
    if (name === 'FIBS Newborn hent') {
        data.BirthTime = data.BirthTime.slice(0, -1)

    }
    this.datasets.set(name, data);
    console.log(name, data);
});

Given('{string} er {string}', function (variable, value) {
    if (value.match(/nytuniktid/i)) {
        value = v4();
    } else if (value.match(/now/i)) {
        value = parseRelativeDate(value);
    } else if (value.match(/nytuniktoid/i)) {
        value = oid();
    } else if (value.match(/patient_cpr/i)) {
        value = patient_cpr;
    } else if (value.match(/number_of_relatives/i)) {
        value = number_of_relatives;
    }

    this.variables.set(variable, value);
    console.log(variable, value);
});

When(
    '{string} kaldes gennem DRG med datasæt {string}',
    { timeout: -1 },
    async function (service, name) {
        console.log(service, name);
        const id = services.get(service);
        if (id == null) {
            throw new Error(`Unknown DRG service id "${service}"`);
        }

        //NAS services kører på dNSP, derfor sættes alternateEndpoint til env_nas. For alle andre services vil alternateEndpoint være undefined, og derved er endpoint lig env.
        let alternateEndpoint;
        if (service.substring(0, 3) == 'NAS') {
            alternateEndpoint = env_nas;
        }

        const body = newBody(id, this.role, alternateEndpoint);
        const dataset = this.datasets.get(name);
        body.fields = dataset;
        //Herunder sættes auth-attributes

        //Rep er specielt i det man skal angive et CPR i auth selvom man ikke har et i body. Vi hardkoder til vi har en løsning på det
        if (dataset.cprNumber) {
            body['auth-attributes'].cpr = dataset.cprNumber;
        }
        if (dataset.CitizenCivilRegistrationNumber) {
            body['auth-attributes'].cpr =
                dataset.CitizenCivilRegistrationNumber;
        }
        if (dataset.PatientCpr) {
            body['auth-attributes'].cpr = dataset.PatientCpr;
        }
        if (dataset.CPR) {
            body['auth-attributes'].cpr = dataset.CPR;
        }
        if (dataset.serviceStartTime) {
            dataset.serviceStartTime = parseRelativeDate(dataset.serviceStartTime).replaceAll(/[^0-9 ]/g, "");
            console.log(dataset.serviceStartTime)
        }
        if (dataset.serviceStopTime) {
            dataset.serviceStopTime = parseRelativeDate(dataset.serviceStopTime).replaceAll(/[^0-9 ]/g, "");
            console.log(dataset.serviceStopTime)
        }
        console.log(body.fields);
        const res = await drgWithAttach(this.attach, body);
        this.responses.push(res);
    }
);

When(
    '{string} kalder {string} gennem DRG med datasæt {string} og svar gemmes som {string}',
    { timeout: -1 },
    async function (role, service, datasetName, responseName) {
        const id = services.get(service);
        role = this.replaceVariables(role);
        if (id == null) {
            throw new Error(`Unknown DRG service id "${service}"`);
        }
        const roleId = roles.get(role);
        if (roleId == null) {
            throw new Error(`Unknown role "${role}"`);
        }

        //NAS services kører på dNSP, derfor sættes alternateEndpoint til env_nas. For alle andre services vil alternateEndpoint være undefined, og derved er endpoint lig env.
        let alternateEndpoint;
        let body;
        if (service.substring(0, 4) == 'FIBS') {
            alternateEndpoint = env_nas;

            body = newBody(id, this.role, alternateEndpoint);
        } else {
            body = newBody(id, roleId)
        }

        const dataset = this.datasets.get(datasetName);
        body.fields = dataset;

        //Herunder sættes auth-attributes

        //Rep er specielt i det man skal angive et CPR i auth selvom man ikke har et i body. Vi hardkoder til vi har en løsning på det
        if (dataset.cprNumber) {
            body['auth-attributes'].cpr = dataset.cprNumber;
        }
        if (dataset.CitizenCivilRegistrationNumber) {
            body['auth-attributes'].cpr =
                dataset.CitizenCivilRegistrationNumber;
        }
        if (dataset.PatientCpr) {
            body['auth-attributes'].cpr = dataset.PatientCpr;
        }
        if (dataset.CPR) {
            body['auth-attributes'].cpr = dataset.CPR;
        }

        const res = await drgWithAttach(this.attach, body);
        this.variables.set(responseName, res);
        this.responses.push(res);
    }
);

Then('modtager service svarer {string}', function (status) {
    switch (status) {
        case 'ok':
            if (this.getLastResponse().body.code != 200) {
                console.log("Faultstring: \n" + JSON.stringify(
                    this.getLastResponse().body.response.main.Envelope.Body.Fault.faultstring, null, 2)
                );
            }
            expect(this.getLastResponse().body.code).to.eql(200);
            break;
        case 'fejl':
            expect(this.getLastResponse().body.code).to.eql(500);
            break;
        default:
            expect(this.getLastResponse().body.code).to.eql(status);
    }
});

Given('bruger er en {string}', function (role) {
    const roleId = roles.get(role);
    if (!roleId) {
        throw new Error(`Unknown role "${role}"`);
    }
    this.role = roleId;
});

Then('retursvar fejltekst er {string}', function (errormsg) {
    // Denne fixture validerer om en given fejltekst i service svarer til den man forventer
    console.log(errormsg);
    expect(
        this.getLastResponse().body.response.main.Envelope.Body.Fault
            .faultstring._text
    ).to.eql(errormsg);
});

Then(
    '{string} + {int} er lig {string}',
    function (oldNumber, number, newNumber) {
        // Funktionen sammenligner 2 tal inkl evt + eller - et antal eks tal1 + 1 er lig tal2 (som er sand hvis tal1=1 og tal2=2)
        oldNumber = this.replaceVariables(oldNumber);
        newNumber = this.replaceVariables(newNumber);
        console.log(oldNumber, newNumber);
        expect(oldNumber + number).to.eql(newNumber);
    }
);

Then('{string} er lig {string}', function (oldNumber, newNumber) {
    oldNumber = this.replaceVariables(oldNumber);
    newNumber = this.replaceVariables(newNumber);
    expect(oldNumber).to.eql(newNumber);
});

Given('{string} er valgt', function (Element) {
    Element = this.replaceVariables(Element);
    expect(Element).not.to.eql(null);
});

Given('teknisk datasæt {string} med felterne', function (name, dataTable) {
    const data = this.parseDataTable(dataTable);

    this.datasets.set(name, data);
});
