const {
    Before,
    BeforeAll,
    After,
    Given,
    Then,
    When,
} = require('@cucumber/cucumber');
const jq = require('node-jq');
const { expect, parseRelativeDate } = require('../util');
const dkihetypecode = new Map(require('../../config/dkihetypecodes.json'));
const dkiheformatcode = new Map(require('../../config/dkiheformatcodes.json'));
const dkiheidentifier = new Map(require('../../config/dkiheidentifiers.json'));

// This fixture expects a register response to contain a specific document type
Then(
    'registersvar indeholder dokument med dokumenttype {string}',
    async function (documenttype) {
        // Getting the DK IHE Type Code by the dokument type
        const typeValue = dkihetypecode.get(documenttype);

        // This urn is the IHE identifier for typecode. NSP uses this to identify document type
        // https://gazelle.ihe.net/XDStarClient/rimihe-templates/rimihe-TypeCode_XDSDocumentEntry.html
        const typeCode = 'urn:uuid:f0306f51-975f-434e-a61c-c59651d33983';

        // Getting the documents from the last response
        let documentlist =
            this.getLastResponse().body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Making result variable
        let res;
        // If the documents is stored in an array
        if (Array.isArray(documentlist)) {
            // Finding the document with the right type code and type value
            res = await jq.run(
                // Selecting document from an array by type code and value
                `.[] | select(.Classification[] | .__classificationScheme == "${typeCode}" and .__nodeRepresentation == "${typeValue}")`,
                documentlist,
                { input: 'json', output: 'json' }
            );
        } else {
            // Finding the document with the right type code and type value
            res = await jq.run(
                // Selecting document by type code and value
                `. | select(.Classification[] | .__classificationScheme == "${typeCode}" and .__nodeRepresentation == "${typeValue}")`,
                documentlist,
                { input: 'json', output: 'json' }
            );
        }

        // Expecting the document to exist and not be equal to nothing
        expect(res, 'Document must exist').to.not.eql('');

        // Pushing the document
        this.documents.push(res);
    }
);

// This fixture expects a register response to contain a document with a specfic format type
Then(
    'registersvar indeholder dokument med formattype {string}',
    async function (formattype) {
        // Getting the format type code by the format type
        const typeValue = dkiheformatcode.get(formattype);
        // This urn is the IHE identifier for typecode. NSP uses this to identify document type
        // https://gazelle.ihe.net/XDStarClient/rimihe-templates/rimihe-FormatCode_XDSDocumentEntry.html
        const typeCode = 'urn:uuid:a09d5840-386c-46f2-b5ad-9c3699a4309d';

        // Getting the documents from the last response
        let documentlist =
            this.getLastResponse().body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Making result variable
        let res;
        // If the documents is stored in an array
        if (Array.isArray(documentlist)) {
            // Finding the document with the right type code and type value
            res = await jq.run(
                // Selecting document from an array by type code and value
                `.[] | select(.Classification[] | .__classificationScheme == "${typeCode}" and .__nodeRepresentation == "${typeValue}")`,
                documentlist,
                { input: 'json', output: 'json' }
            );
        } else {
            // Finding the document with the right type code and type value
            res = await jq.run(
                // Selecting document by type code and value
                `. | select(.Classification[] | .__classificationScheme == "${typeCode}" and .__nodeRepresentation == "${typeValue}")`,
                documentlist,
                { input: 'json', output: 'json' }
            );
        }

        // Expecting the document to exist and not be equal to nothing
        expect(res, 'Document must exist').to.not.eql('');
    }
);

// This fixture expects a document type to have the right service start time
Then(
    'dokumenttype har service start tidspunkt {string}',
    async function (serviceStartTime) {
        // Getting the last document
        let document = this.getLastDocument();

        // Finding place in document where service start time is
        let res = await jq.run(
            // Selecting from the slots array in the document
            `.Slot[] | select(.ValueList.Value._text? == "${serviceStartTime}")`,
            document,
            { input: 'json', output: 'json' }
        );

        // Expects the result not to be nothing
        expect(res, 'Service start time must exist').to.not.eql('');
    }
);

// This fixture expects the document to have the right service stop time
Then(
    'dokumenttype har service stop tidspunkt {string}',
    async function (serviceStopTime) {
        // Getting the last document
        let document = this.getLastDocument();

        // Finding place in document where service stop time is
        let res = await jq.run(
            // Selecting from the slots array in the document
            `.Slot[] | select(.ValueList.Value._text? == "${serviceStopTime}")`,
            document,
            { input: 'json', output: 'json' }
        );

        // Expects the result not to be nothing
        expect(res, 'Service stop time must exist').to.not.eql('');
    }
);

// This fixture expects the document is made on the creation time
Then(
    'dokumenttype har creation tidspunkt {string}',
    async function (creationTime) {
        // Getting the last document
        let document = this.getLastDocument();

        // Finding place in document where creation time is
        let res = await jq.run(
            // Selecting from the slots array in the document
            `.Slot[] | select(.ValueList.Value._text? == "${creationTime}")`,
            document,
            { input: 'json', output: 'json' }
        );

        // Expects the result not to be nothing
        expect(res, 'Creation time must exist').to.not.eql('');
    }
);

// This fixture expects the document to have the right author person
Then('dokumenttype har author person {string}', async function (authorPerson) {
    // Splitting the name from the fixture
    let name = authorPerson.split(' ');

    // First element in array is the first name
    let firstname = name[0];

    // Second element in the array is the last name
    let lastname = name[1];

    // Getting the last document
    let document = this.getLastDocument();

    if (firstname == 'Morten') {
        console.log(JSON.stringify(document, null, 2));
    }

    let res;
    expect(res, 'Author person must exist').to.not.eql('');
});

// This fixture expects a document to have the id
Then(
    '{string} dokumentet har id {string}',
    function (documenttype, documentid) {
        // Expects document id from the last document to equal document id from fixture
        expect(this.getLastDocument().__id).to.eql(documentid);
    }
);

// This fixture is based on predefined data and this patient CPR contains the data that makes the testcase pass
Given('{string} har alle datatyper i stamkort', function (cpr) {
    cpr = this.replaceVariables(cpr);
    // Expects CPR to be specific CPR
    expect(cpr).to.eql('1603900967');
});

// This fixture expects the CDA document to contain an amount of components
Then('CDA indeholder {int} komponenter i alt', function (countElements) {
    // Getting components from the last response
    const componentlist =
        this.getLastResponse().body.response.attachments[0].ClinicalDocument
            .component.structuredBody.component.section.entry;

    // Expecting the component list to contain the given number of components
    expect(componentlist.length).to.eql(countElements);
});

// This fixture expects the CDA document to contain an amount of a specific component type
Then(
    'CDA indeholder {int} komponenter af typen {string}',
    async function (countelements, componenttype) {
        // Finding the documents with an specific component type in the last response
        const componententries = await jq.run(
            `[.[]|select(.observation.code|select(."__code"? =="${componenttype}"))]`,
            this.getLastResponse().body.response.attachments[0].ClinicalDocument
                .component.structuredBody.component.section.entry,
            { input: 'json', output: 'json' }
        );

        // Expecting the component list to contain the given number of components
        expect(componententries.length).to.eql(countelements);
    }
);

// This fixture expects the CDA document observation to contain value and verifies that the individual observations fx LivingWill has the specified value
Then(
    'CDA observation {string} indeholder {string}',
    async function (componenttype, value) {
        // Finding the documents with an specific component type in the last response
        const componententries = await jq.run(
            `[.[]|select(.observation.code|select(."__code"? =="${componenttype}"))]`,
            this.getLastResponse().body.response.attachments[0].ClinicalDocument
                .component.structuredBody.component.section.entry,
            { input: 'json', output: 'json' }
        );

        // Expecting the first component's observation value to equal the value from the fixture
        expect(componententries[0].observation.value.__extension).to.eql(value);
    }
);

// This fixture saves home community id, repository id and document unique id from a document
Then(
    'dokumentoplysningerne fra dokument {string} gemmes som {string}, {string} og {string}',
    async function (
        DKIHETypeCode,
        HomeCommunityID,
        RepositoryUniqueID,
        DocumentUniqueID
    ) {
        // Getting documents from last response
        let documentlist =
            this.getLastResponse().body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Converting document type to type code
        const dkIHETypeCode = dkihetypecode.get(DKIHETypeCode);

        // Finding document with type code
        const document = await jq.run(
            `.[] | select(.Classification[] | .__nodeRepresentation == "${dkIHETypeCode}")`,
            documentlist,
            { input: 'json', output: 'json' }
        );

        // Saving the ids into variables
        const homeCommunityId = document.__home;
        const documentUniqueId = document.__id;
        const repositoryUniqueId = documentUniqueId.substring(
            0,
            documentUniqueId.indexOf('^')
        );

        // Setting the id's
        this.variables.set(HomeCommunityID, homeCommunityId);
        this.variables.set(RepositoryUniqueID, repositoryUniqueId);
        this.variables.set(DocumentUniqueID, documentUniqueId);

        // Expecting the document id no to be null
        expect(this.documentId).to.not.be.null;
        expect(this.documentHome).to.not.be.null;
    }
);

// This fixture expects the CDA document to contain doctor with name, yderNr and postal code
Then(
    'CDA indeholder læge med navn {string} og ydernr {string} og postnr {string}',
    function (providerName, providerNumber, postalCode) {
        // Getting health provider from last response
        const provider =
            this.getLastResponse().body.response.attachments[0].ClinicalDocument
                .recordTarget.patientRole.providerOrganization;

        // Expecting name, yderNr and postal code from repsonse to equal name, yderNr and postal code from fixture
        expect(provider.name._text).to.eql(providerName);
        expect(provider.id.__extension).to.eql(providerNumber);
        expect(provider.addr.postalCode._text).to.eql(postalCode);
    }
);

// This fixture counts the number of documents in the response and saves the amount
Then(
    'antallet af dokumenter i {string} gemmes som {string}',
    function (responseName, resultName) {
        // Getting response
        const response = this.variables.get(responseName);

        // Expecting response not to be null
        if (response == null) {
            throw new Error('Unable to find response: ' + responseName);
        }

        // Getting the documents from the response
        let documentlist =
            response.body.response.main.Envelope.Body.AdhocQueryResponse
                .RegistryObjectList.ExtrinsicObject;

        // If the documents are in an array
        if (Array.isArray(documentlist)) {
            // Setting result name to the number of documents
            this.variables.set(resultName, documentlist.length);

            // If the document list is undefined
        } else if (documentlist == undefined) {
            // Setting the result name to zero
            this.variables.set(resultName, 0);
        } else {
            // Setting the result name to one
            this.variables.set(resultName, 1);
        }
    }
);

// This fixture expects the respitory to contain warning
Then(
    'repository {string} indeholder en advarsel om {string}',
    function (ResponseName, Error) {
        // Getting warning
        const error =
            this.variables.get(ResponseName).body.response.main.Envelope.Body
                .RetrieveDocumentSetResponse.RegistryResponse.RegistryErrorList
                .RegistryError.__codeContext;

        // Expecting the warning to be the same as in the fixture
        expect(error).to.eql(Error);
    }
);

// This fixture finds a document with a specific document id and pushes it
Then('dokumentet med {string} findes', async function (documentUniqueID) {
    // Getting documents from last response
    let documentlist =
        this.getLastResponse().body.response.main.Envelope.Body
            .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

    // Replacing document unique id
    documentUniqueID = this.replaceVariables(documentUniqueID);

    // Finding document with type code
    const document = await jq.run(
        `.[]  | select(.ExternalIdentifier[] | select(.__value == "${documentUniqueID}"))`,
        documentlist,
        { input: 'json', output: 'json' }
    );

    // Attaching document
    // this.attach(JSON.stringify(document, null, 2), 'text/plain');
    // Pushing document
    this.documents.push(document);

    // Expecting the document to be not empty
    expect(document).to.not.eql("")
});

// This fixture expects a metadata field to have a given value with a specific codesystem code
Then(
    'metadatafelt {string} har værdi {string} og codesystem {string}',
    async function (field, value, codesystem) {
        // Getting last document
        const document = this.getLastDocument();

        // Getting DK IHE identifier urn uuid
        const identifierId = dkiheidentifier.get(field);

        // Finding classification with identifier id
        const classification = await jq.run(
            `.Classification[]|select(.__classificationScheme == "${identifierId}")`,
            document,
            {
                input: 'json',
                output: 'json',
            }
        );

        expect(classification.Slot.ValueList.Value._text).to.eql(codesystem);
        expect(classification.Name.LocalizedString.__value).to.eql(value);
    }
);

// This fixture expects a metadata field to equal a given value
Then('metadatafelt {string} har værdien {string}', function (field, value) {
    // Getting last document
    const document = this.getLastDocument();

    // Adding underscores to field
    field = '__' + field;

    // Expecting field to equal the value
    expect(document[field]).to.eql(value);
});

// This fixture expects the slot array to have fields with values
Then(
    'slot metadatafelt {string} har værdien {string}',
    async function (field, value) {
        // Getting last document
        const document = this.getLastDocument();

        // Getting element from slot array
        const element = await jq.run(
            `.Slot[]|select(.__name == "${field}")`,
            document,
            { input: 'json', output: 'json' }
        );
        
        if (value == 'Now') {
            value = parseInt(parseRelativeDate(value).replace(/[^0-9]/g, ""))
            expect(parseInt(element.ValueList.Value._text)).to.be.within(
                value - 2, value + 2
            );
            return;
        }

        expect(element.ValueList.Value._text).to.eql(value);
    }
);

// This fixture expects a metadata field to exist
Then('metadatafelt {string} findes', async function (field) {
    // Getting last document
    const document = this.getLastDocument();

    // Getting array of booleans
    const element = await jq.run(`[.Slot[].__name == "${field}"]`, document, {
        input: 'json',
        output: 'json',
    });

    // Expects one of the elements in the array to be true
    expect(element.includes(true)).to.be.true;
});

// This fixture expects the patient id to have a given value
Then(
    'metadatafelt patient id {string} har værdien {string}',
    async function (field, value) {
        // Getting last document
        const document = this.getLastDocument();

        // Replacing variables
        value = this.replaceVariables(value);

        // Getting DK IHE identifier urn uuid
        const identifierId = dkiheidentifier.get(field);

        // Getting element with patient id
        const element = await jq.run(
            `.ExternalIdentifier[] | select(.__identificationScheme == "${identifierId}")`,
            document,
            { input: 'json', output: 'json' }
        );

        expect(element.__value).to.eql(value);
    }
);

// This fixture expects the source patient info to have 3 values
Then(
    'metadatafelt {string} har værdierne {string}, {string} og {string}',
    async function (field, value1, value2, value3) {
        // Getting last document
        const document = this.getLastDocument();

        // Getting element from slot array
        const element = await jq.run(
            `.Slot[]|select(.__name == "${field}")`,
            document,
            { input: 'json', output: 'json' }
        );

        expect(element.ValueList.Value[0]._text.substring(0, 5)).to.eql(value1);
        expect(element.ValueList.Value[1]._text.substring(0, 5)).to.eql(value2);
        expect(element.ValueList.Value[2]._text.substring(0, 5)).to.eql(value3);
    }
);

Then(
    'dokumenttype har metadatafelt {string} med værdien {string}',
    async function (dkiIdentifier, healthcareFacilityTypeCode) {
        //Getting last document
        const document = this.getLastDocument();

        // Getting DK IHE identifier urn uuid
        const identifierId = dkiheidentifier.get(dkiIdentifier);

        // Getting element with patient id
        const element = await jq.run(
            `.Classification[] |select(.__classificationScheme=="${identifierId}")`,
            document,
            { input: 'json', output: 'json' }
        );

        expect(element.__nodeRepresentation).to.eql(healthcareFacilityTypeCode);
    }
);
