const {
    Before,
    BeforeAll,
    After,
    Given,
    Then,
    When,
} = require('@cucumber/cucumber');
const jq = require('node-jq');
const { expect } = require('../util');
const { getDDSRegistryWithPoces } = require('../../dataaccess/ddsregistry');
const { personLookup } = require('../../dataaccess/sces');
let dayjs = require('dayjs');
dayjs().format();
const { appointmentXML } = require('../../config/appointment');

/* DK-IHE Type Codes: https://view.officeapps.live.com/op/view.aspx?src=https%3A%2F%2Fsvn.medcom.dk%2Fsvn%2Fdrafts%2FStandarder%2FIHE%2FOID%2FDK-IHE_Metadata-Common_Code_systems-Value_sets.xlsx&wdOrigin=BROWSELINK */
const dkihetypecodes = new Map(require('../../config/dkihetypecodes.json'));

// This fixture expects the DROS status to equal the status from the fixture
Then('DROS status er {string}', function (Status) {
    // Getting the status from the last response
    let response =
        this.getLastResponse().body.response.main.Envelope.Body.RegistryResponse
            .__status;

    // Expecting the status from the response status to equal the status from the fixture
    // Note: The status is either 'Succes' or 'Failure', which both are 7 letters. Therefore this fixture only works when the status is 7 letters
    expect(response.substring(response.length - 7)).to.equal(Status);
});

// This fixture expects the DROS failure message to equal the given failure message
Then('DROS fejltekst er {string}', function (failureMessage) {
    // Getting the failure message from the last response
    let failureResponse =
        this.getLastResponse().body.response.main.Envelope.Body.RegistryResponse
            .RegistryErrorList.RegistryError.__codeContext;

    // Formatted failure message by splitting it
    let failureResponseEdited = failureResponse.split(' UUID');

    // Expecting the failure message from the response to equal the message from the fixture
    expect(failureResponseEdited[0]).to.equal(failureMessage);
});

// This fixture gets all the DDS Registry documents and saves them
Given(
    'DDS registry dokumenter hentes for {string} og gemmes som {string}',
    { timeout: -1 },
    async function (PatientCPR, ResultName) {
        // Convert CPR
        PatientCPR = this.replaceVariables(PatientCPR);

        // Getting the DDS Registry documents with POCES
        const result = await getDDSRegistryWithPoces(this.attach, PatientCPR);

        // Setting the ResultName variable with the DDS Registry response
        this.variables.set(ResultName, result);
    }
);

// This fixture counts the DDS documents in a given response and saves the number of documents
Given(
    'antal DDS dokumenter i {string} gemmes som {string}',
    function (ResponseName, NumberOfDocuments) {
        // Getting the documents from the ResponseName given in the fixture
        const extrinsicObject =
            this.variables.get(ResponseName).body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Note: If there are more than 1 document, the documents will be stored in an array. If there is only 1 document, it won't be stored in an array.
        // If there is more than 1 document, therefore an array
        if (Array.isArray(extrinsicObject)) {
            // Setting the number of documents
            this.variables.set(NumberOfDocuments, extrinsicObject.length);
            // Attaching to see the number of documents in Jenkins
            this.attach('Antal dokumenter: ' + extrinsicObject.length);
        } else if (extrinsicObject != undefined) {
            // Setting the number of documents to 1
            this.variables.set(NumberOfDocuments, 1);
            // Attaching to see the number of documents in Jenkins
            this.attach('Antal dokumenter: ' + 1);
        }
    }
);

// This fixture expects a response to contain a document with a specific id
Then(
    'dokumentet med {string} findes i {string}',
    async function (Id, ResponseName) {
        // Replace id
        const id = this.replaceVariables(Id);

        // Getting the documents from the response name
        const extrinsicObjects =
            this.variables.get(ResponseName).body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Making variable
        let extrinsicObject;

        // Note: If there are more than 1 document, the documents will be stored in an array. If there is only 1 document, it won't be stored in an array.
        // If there is more than 1 document, therefore an array
        if (Array.isArray(extrinsicObjects)) {
            // Finding the document with the specific id
            extrinsicObject = await jq.run(
                // Selecting from an array of documents, then from another array with a value that ends with the id
                `.[]|select(.ExternalIdentifier[]|select(.__value | endswith("${id}")))`,
                extrinsicObjects,
                { input: 'json', output: 'json' }
            );
        } else if (extrinsicObjects != undefined) {
            // It is not an array and not undefined, therefore 1 document
            // Finding the document with the specific id
            extrinsicObject = await jq.run(
                // Selectiong from an array a value that ends with the id
                `.ExternalIdentifier[]|select(.__value | endswith("${id}"))`,
                extrinsicObjects,
                { input: 'json', output: 'json' }
            );
        }

        // Expecting the document not to be undfined
        expect(extrinsicObject).not.to.equal(undefined);
    }
);

// This fixture finds a specific document type in a response and saving the document id
Given(
    '{string} dokument findes i {string} og dokumentid gemmes som {string}',
    async function (DKIHETypeCode, ResponseName, Id) {
        // Converting the document name to a DK IHE Type Code
        const dkIHETypeCode = dkihetypecodes.get(DKIHETypeCode);

        // Getting the documents from the response name
        const response =
            this.variables.get(ResponseName).body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Selecting the document(s) with the right type code from an array of documents, and saving the documents in an array
        let extrinsicObject = await jq.run(
            `[.[]|select(.Classification[]|select(.__nodeRepresentation=="${dkIHETypeCode}"))]`,
            response,
            { input: 'json', output: 'json' }
        );

        // Setting the id variable to the first document id in the array
        this.variables.set(Id, extrinsicObject[0].__id.substring(9));
        // Using substring because the __id contains more than the id
    }
);

// This fixture expects a response name to contain a document with a field with the value and id
Then(
    'feltet {string} er lig værdien {string} i dokumentet med {string} i {string}',
    async function (Field, Value, Id, ResponseName) {
        // Replace id
        const id = this.replaceVariables(Id);

        // Getting the documents from the response name
        const response =
            this.variables.get(ResponseName).body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Returning true or false according to if the document exists
        let isExisting = await jq.run(
            `if(.[]|select((.Slot[]|select(.__name=="${Field}" and .ValueList.Value._text == "${Value}")) and (.ExternalIdentifier[]|select(.__value | endswith("${id}"))))) then (true) else (false) end`,
            response,
            { input: 'json', output: 'json' }
        );

        // Expects the document to exist - the boolean to be true
        expect(isExisting).to.be.true;
    }
);

// This fixture expects a response name to contain a document with a specific document type code
Then(
    'dokumenttype for {string} i {string} er {string}',
    async function (Id, ResponseName, DocumentType) {
        // Replace id
        const id = this.replaceVariables(Id);

        // Converting the document name to a DK IHE Type Code
        const dkIHETypeCode = dkihetypecodes.get(DocumentType);

        // Getting the documents from the response name'
        const response =
            this.variables.get(ResponseName).body.response.main.Envelope.Body
                .AdhocQueryResponse.RegistryObjectList.ExtrinsicObject;

        // Returning true or false according to if the document exists
        const isExisting = await jq.run(
            `if(.[]|select((.Classification[]|select(.__nodeRepresentation == "${dkIHETypeCode}")) and (.ExternalIdentifier[]|select(.__value | endswith("${id}"))))) then (true) else (false) end`,
            response,
            { input: 'json', output: 'json' }
        );

        expect(isExisting).to.be.true;
    }
);

// This fixture makes an appointment document with values from a data table
Given(
    'der laves et aftaledokument {string} for {string} med værdierne',
    async function (Document, PatientCPR, dataTable) {
        // Replacing CPR
        PatientCPR = this.replaceVariables(PatientCPR);

        // Parsing data table
        const data = this.parseDataTable(dataTable);

        // Looking up person to get personal information
        const response = await personLookup(this.attach, PatientCPR);

        // Getting patient information from the response
        const patientInformation =
            response.body.response.main.Envelope.Body.getPersonInformationOut
                .PersonInformationStructure;

        // Setting street name in data table
        data.streetname =
            patientInformation.PersonAddressStructure.AddressComplete.AddressPostal.StreetName._text;

        // Setting district name in data table
        data.districtName =
            patientInformation.PersonAddressStructure.AddressComplete.AddressPostal.DistrictName._text;

        // Setting post code in data table
        data.postCodeIdentifier =
            patientInformation.PersonAddressStructure.AddressComplete.AddressPostal.PostCodeIdentifier._text;

        // Setting street building in data table
        data.streetBuildingIdentifier =
            patientInformation.PersonAddressStructure.AddressComplete.AddressPostal.StreetBuildingIdentifier._text;

        // Setting gender in data table
        data.personGenderCode =
            patientInformation.RegularCPRPerson.PersonGenderCode._text;

        // Setting first name in data table
        data.firstName =
            patientInformation.RegularCPRPerson.SimpleCPRPerson.PersonNameStructure.PersonGivenName._text;

        // Setting last name in data table
        data.lastName =
            patientInformation.RegularCPRPerson.SimpleCPRPerson.PersonNameStructure.PersonSurnameName._text;

        // Setting birth date in data table
        data.birthDate =
            patientInformation.RegularCPRPerson.PersonBirthDateStructure.BirthDate._text
                .substring(0, 10)
                // Replacin - in birth date with nothing
                .replace(/-/gi, '');

        // Setting date in data table to today
        data.date = dayjs(new Date()).format('DD-MM-YYYY');

        // Setting document variable with the appointment document made by using data table and cpr
        this.variables.set(Document, appointmentXML(PatientCPR, data));

        // Attaching the appointment document, so it is visible in Jenkins
        this.attach(appointmentXML(PatientCPR, data));
    }
);

// This fixtures base 64 encodes the document and saves the encoded document
Given(
    '{string} base64encodes og gemmes som {string}',
    function (Document, ResultName) {
        // Getting the document
        const document = this.variables.get(Document);

        // Using Buffer function to encode the document
        const encodedDocument = Buffer.from(document).toString('base64');

        // Setting the result name variable to the encoded document
        this.variables.set(ResultName, encodedDocument);
    }
);
