const fs = require('fs/promises');
const readline = require('node:readline/promises');
const { stdin: input, stdout: output } = require('node:process');
const { getWsdl } = require('../../dataaccess/util');

const fetchWsdlFromUrlAsText = async (wsdlUrl) => {
    console.log(`Fetching from ${wsdlUrl}`);
    return (await getWsdl(wsdlUrl)).text;
};

const extractSchemaLocationUrlsFromText = (text) => {
    const regex = /schemaLocation="(.*?)"/g;
    const matches = [...text.matchAll(regex)];
    const locations = matches.map(match => {
        return match[1];
    });
    console.log('Found schemaLocations:', locations);
    return locations;
};

const fetchOneWsdlAndExtractSubSchemaLocations = async (url) => {
    const wsdlText = await fetchWsdlFromUrlAsText(url);
    const subSchemaLocations = extractSchemaLocationUrlsFromText(wsdlText);
    const resolvedSubSchemaLocations = subSchemaLocations.map(subSchemaLocation => {
        const resolvedUrl = new URL(subSchemaLocation, url); // Resolve full URL in case of relative URLs in subSchemaLocation. Like "../IHE/xcf.xsd". Yes, that happens..
        return resolvedUrl.href;
    });
    console.log('');

    return {
        wsdlText,
        subSchemaLocations: resolvedSubSchemaLocations
    };
};

const fetchManyWsdl = async (baseWsdlUrl, readlineInterface) => {
    const schemaLocationsStack = [];
    schemaLocationsStack.push(baseWsdlUrl);

    const resultMap = {};
   
    while (schemaLocationsStack.length !== 0) {
        const nextSchemaLocation = schemaLocationsStack.pop();
        const { wsdlText, subSchemaLocations } = await fetchOneWsdlAndExtractSubSchemaLocations(nextSchemaLocation);
        resultMap[nextSchemaLocation] = { wsdlText };

        for (const subSchemaLocation of subSchemaLocations) {
            if (!resultMap[subSchemaLocation] && !schemaLocationsStack.includes(subSchemaLocation)) {
                const addThis = await readlineInterface.question(`Do you want to fetch from ${subSchemaLocation} ? [y / n]: `);

                if (addThis === 'y') {
                    schemaLocationsStack.push(subSchemaLocation);
                }
            }
        }
    }

    return resultMap;
};

const addBestName = (urlsAndFiles) => {
    const urls = Object.keys(urlsAndFiles);

    // find length of longest common prefix
    let i = 0;
    while (urls[0][i] && urls.every(w => w[i] === urls[0][i])) {
        i++;
    }
    
    for (const [key, value] of Object.entries(urlsAndFiles)) {
        const strippedUrl = key.substring(i);
        const replaced = strippedUrl.replace(/([^a-zA-Z0-9]+)/g, '_'); // do some defensive filename sanitation
        value.bestName = replaced;
    }
};

const saveMany = async (locationWsdlMap, readlineInterface) => {
    console.log('');
    console.log('We will now save the files to a folder.');
    const folderName = await readlineInterface.question('What folder do you want to save the files in? Make sure that it exists already. Example "brs/test1" ');
    const prefix = await readlineInterface.question('What prefix do you want for each file? Example "brs" ');
    const postfix = await readlineInterface.question('What postfix do you want for each file? Example "test2" ');
    console.log('');
    console.log('You will now be asked for a short name for each file.')
    console.log(`The files will be saved to ${__dirname}/known-wsdl/${folderName}/${prefix}-<name>-${postfix}`);
    console.log('Each name has a suggestion. If you press <enter> with blank input, the suggestion will be used.');

    const urlsAndFiles = [];

    for (const [key, value] of Object.entries(locationWsdlMap)) {
        console.log('');
        console.log(`What name should the WSDL from ${key} be saved as?`);
        const name = (await readlineInterface.question(`Suggestion is "${value.bestName}". Press <enter> to accept or write you own name: `)) || value.bestName;
        const completeName = `${folderName}/${prefix}-${name}-${postfix}`
        await fs.writeFile(`${__dirname}/known-wsdl/${completeName}`, value.wsdlText);
        urlsAndFiles.push({
            url: key,
            file: completeName
        });
    }

    return urlsAndFiles;
};

const printUrlsAndFiles = (urlsAndFiles) => {
    let maxUrlLength = 0;
    let maxFileLength = 0;
    for (const urlAndFile of urlsAndFiles) {
        const urlLength = urlAndFile.url.length;
        const fileLength = urlAndFile.file.length;
        if (maxUrlLength < urlLength) {
            maxUrlLength = urlLength;
        }
        if (maxFileLength < fileLength) {
            maxFileLength = fileLength;
        }
    }

    const makePairString = (url, file) => {
        const urlPaddingAmount = maxUrlLength - url.length;
        const filePaddingAmount = maxFileLength - file.length;
        return `| ${url}${' '.repeat(urlPaddingAmount)} | ${file}${' '.repeat(filePaddingAmount)} |`;
    };

    console.log('');
    console.log(makePairString('wsdlUrl', 'filNavn'));
    for (const urlAndFile of urlsAndFiles) {
        console.log(makePairString(urlAndFile.url, urlAndFile.file));
    }
    console.log('');
};

const wsdlFetcherMain = async () => {
    const readlineInterface = readline.createInterface({ input, output });

    console.log('');
    console.log('Welcome to the WSDL-fetcher! Together, we will fetch some good stuff.');
    console.log('');

    const baseWsdlUrl = await readlineInterface.question('Paste URL for base-WSDL-file:\n');
    const locationWsdlMap = await fetchManyWsdl(baseWsdlUrl, readlineInterface);

    console.log('All locations found: ', Object.keys(locationWsdlMap));

    addBestName(locationWsdlMap);

    const urlsAndFiles = await saveMany(locationWsdlMap, readlineInterface);

    printUrlsAndFiles(urlsAndFiles);

    readlineInterface.close();
    return;
};

if (require.main === module) {
    wsdlFetcherMain();
}
